package chase.s2;

import java.awt.geom.Point2D;

import robocode.util.Utils;

public final class Tools {
	private final static double oos2p = (1.0/Math.sqrt(2.0*Math.PI));
	public static final double gaussian(double u) {
		return oos2p*Math.pow(Math.E,-0.5*u*u);
	}
	public static final double kernel(double[] data, int index, double bandwidth) {
		int n = data.length;
		double x = data[index];
		double k = 0;
		for(int i=1;i<n;++i)
			k += gaussian((x - data[i])/bandwidth);
		k *= 1.0/(n*bandwidth);
		return 0;
	}
	public static final int sign(double input) {
		return input < 0.0 ? -1 : 1;
	}
	public static final double rollingAvg(double value, double newEntry, double depth, double weighting ) {
		return (value * depth + newEntry * weighting)/(depth + weighting);
	}
	public final static double limit(double min, double value, double max) {
		return Math.max(min, Math.min(value, max));
	}

	public final static Point2D.Double project(Point2D p, double ang, double dist) {
		return new Point2D.Double(p.getX() + dist*Math.sin(ang), p.getY() + dist*Math.cos(ang));
	}
	
	public static final double absoluteAngle(Point2D source, Point2D target) {
		return Math.atan2(target.getX() - source.getX(), target.getY() - source.getY());
	}
	public static final double botWidth(double absAngle) {
		final double n = 16.568542494923801952067548968388;
		double v = Math.abs(Math.sin(Utils.normalRelativeAngle(absAngle)*2.0));
		return v * n + 40;
	}

	/**
	 * That actual collide-able angle
	 */
	public static final double botWidthAngle(double absAngle, double distance) {
		final double w = botWidth(absAngle);
		return Math.atan(w / (distance + (w - 56)));
	}

	public static double timeWeightedAverage(double newValue, double oldValue,
			double deltaTime, double decayRate) {
		double w = Math.exp(-decayRate * deltaTime);
		return (w * oldValue + (1 - w) * newValue);
	}

	public static double computeDecayRate(double halfLife) {
		return (-Math.log(0.5) / halfLife);
	}
}
