package chase.s2.gun;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import robocode.Rules;
import robocode.ScannedRobotEvent;
import robocode.util.Utils;
import chase.s2.Extension;
import chase.s2.Genesis;
import chase.s2.Tools;

public class GunMelee extends Extension {
	HashMap<String,EnemyData> enemyData = new HashMap<String,EnemyData>(); 
	
	static final class EnemyData {
		double lastEnergy = 0;
		double velocity = 0;
		double heading = 0;
		Point2D location;
	}
	
	static final class AimData {
		double angle = 0;
		double distance = 0;
		double angleWidth = 0;
	}
	
	double bulletPower = 0;
	boolean hasTarget = false;
	
	public void execute() {
		if(Genesis.isChallenge) return;
		if(bot.getOthers() < 2) return;
		Point2D myLocation = new Point2D.Double(bot.getX(), bot.getY());
		
		if(bot.getGunHeat() < 0.001 && hasTarget
			&& bot.getGunTurnRemainingRadians() < Math.PI/128) {
			bot.setFire(bulletPower);
		}
		
		if(bot.getGunHeat() / bot.getGunCoolingRate() > 5) {
			hasTarget = false;
			return;
		}
		
		//Calculate angles!
		Iterator<EnemyData> it = enemyData.values().iterator();
		
		double nearest = Double.POSITIVE_INFINITY;
		while(it.hasNext()) {
			double dist = it.next().location.distance(myLocation);
			if(dist < nearest)
				nearest = dist;
		}
		
		
		bulletPower = 400.0 / nearest;
		
		it = enemyData.values().iterator();
		List<AimData> aData = new ArrayList<AimData>();
		while(it.hasNext()) {
			EnemyData data = it.next();
			AimData ad = new AimData();
			ad.distance =  data.location.distance(myLocation);
			ad.angle = Tools.absoluteAngle(myLocation, data.location);
			ad.angleWidth = Tools.botWidthAngle(ad.angle, ad.distance) / 2.0;
			if(nearest < 200 && Math.abs(ad.distance-nearest) < 0.1) {
				hasTarget = true;
				
				bulletPower = 2.4;
				if(ad.distance < 150)
					bulletPower = 3;
				
//				double bft = ad.distance / Rules.getBulletSpeed(bulletPower);
//				Point2D.Double n = Tools.project(data.location, data.heading, data.velocity*bft);
//				ad.angle = Tools.absoluteAngle(myLocation, n);
//				bot.setTurnGunRightRadians(Utils.normalRelativeAngle(ad.angle - bot.getGunHeadingRadians()));
				
				Point2D target = data.location;
				double bSpeed = 20 - (3 * bulletPower);
				int tickCounter = 0;
				while(myLocation.distance(target) >= bSpeed*tickCounter++
						&& battlefield.contains(target)) {
					target = Tools.project(target, data.heading, data.velocity);
				}
				
				ad.angle = Tools.absoluteAngle(myLocation, target);
				bot.setTurnGunRightRadians(Utils.normalRelativeAngle(ad.angle - bot.getGunHeadingRadians()));
				return;
			}
			aData.add(ad);
			ad = new AimData();
			//Calculate where they could be by the time our bullet gets to them
			double bft = ad.distance / Rules.getBulletSpeed(bulletPower);
			Point2D.Double n = Tools.project(data.location, data.heading, data.velocity*bft);
			//Limit by the wall
			if(!battlefield.contains(n)) {
				if(n.x > battlefield.getMaxX()) n.x = battlefield.getMaxX();
				if(n.x < battlefield.getMinX()) n.x = battlefield.getMinX();
				if(n.y > battlefield.getMaxY()) n.y = battlefield.getMaxY();
				if(n.y > battlefield.getMinY()) n.y = battlefield.getMinY();
			}
			ad.distance =  n.distance(myLocation);
			ad.angle = Tools.absoluteAngle(myLocation, n);
			ad.angleWidth = Tools.botWidthAngle(ad.angle, ad.distance) / 2.0;
			aData.add(ad);
			
		}
		hasTarget = false;
		
		//If I knew an easier way to do this, I would use it
		double[] bin = new double[360];
		for(AimData d : aData) {
			double[] angles = new double[5];
			for(int i=-2; i<3; ++i) {
				double offset = i / 2.0;
				angles[i+2] = Math.toDegrees(Utils.normalAbsoluteAngle(d.angle + offset * d.angleWidth));
				double weight = 1/Math.max(Math.sqrt(d.distance),14);
				for(int j=0; j<bin.length; ++j) {
					double amount = Math.abs(j - angles[i+2]);
					if(amount == 0) bin[j] += weight;
					if(amount > 0 && amount < 1)
						bin[j] += amount * weight;
				}
			}
		}
		
		//Determine bin scores!
		double[] scores = new double[bin.length];
		for(int i=0; i<bin.length; ++i) {
			double score = 0;
			//It is a little complicated in there, but things will
			//   never go out of bounds
			for(int j=-1; j<1+1; ++j) {
				int abs = Math.abs(j) + 1;
				if(i+j < 0 || i+j >= bin.length) {
					if(i == 0 || i == bin.length - 1) {
						score += bin[i - j] / abs;
					} else {
						int offset = (j < 0) ? -3 : 3;
						score += bin[i - j + offset] / abs;
					}
				} else {
					score += bin[i + j] / abs;
				}
			}
			scores[i] = score;
		}
		
		//Determine the best angle
		int bestIndex = (bin.length-1)/2; /* TODO: This could be problematic */
		for(int i=0; i<bin.length; ++i)
			if(scores[i] > scores[bestIndex])
				bestIndex = i;
		
		hasTarget = true;
		double aimAngle = Math.toRadians(bestIndex);
		bot.setTurnGunRightRadians(Utils.normalRelativeAngle(aimAngle - bot.getGunHeadingRadians()));
	}
	
	public void onScannedRobot(ScannedRobotEvent e) {
		if(Genesis.isChallenge) return;
		if(bot.getOthers() < 2) return;
		
		EnemyData data = enemyData.get(e.getName());
		if(data == null) data = new EnemyData();
		data.lastEnergy = e.getEnergy();
		
		double eDistance = e.getDistance();
		double eAbsAngle = bot.getHeadingRadians() + e.getBearingRadians();
		Point2D myLocation = new Point2D.Double(bot.getX(), bot.getY());
		Point2D eLocation = Tools.project(myLocation, eAbsAngle, eDistance);
		
		data.heading = e.getHeadingRadians();
		data.velocity = e.getVelocity();
		data.location = eLocation;
		enemyData.put(e.getName(), data);
		
		//That is all taken care of!
	}
}
