package chase.s2.move;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import robocode.BulletHitEvent;
import robocode.Rules;
import robocode.ScannedRobotEvent;
import robocode.util.Utils;
import chase.s2.Extension;
import chase.s2.Genesis;
import chase.s2.Tools;

public class APMM extends Extension {
	private static final int MAX_MATCH_CHECKS = 400;
	public static final HashMap<Integer, State> map = new HashMap<Integer, State>();
	public static State head;
	ArrayList<Wave> waves = new ArrayList<Wave>();
	
	Point2D.Double myLastLocation;
	
	HashMap<String,EnemyData> enemyData = new HashMap<String,EnemyData>(); 
	
	public static double lHeading, lVelocity;
	
	static final class EnemyData {
		double lastEnergy = 0;
		Point2D location;
	}
	
	public final void execute() {
		if(bot.getTime() % (10 + (int)(Math.random()*40)) == 0) {
			myLastLocation = new Point2D.Double(bot.getX(), bot.getY());			
		}
	}
	
	public final void run() {
		addBreak();
		lVelocity = 100;
	}
	
	Point2D myLocation;
	
	public final void onScannedRobot(ScannedRobotEvent e) {
		if (Genesis.isReference) return;
		if (bot.getEnergy() < 0.1)
			return;
		double eHeading = bot.getHeadingRadians();
		double eVelocity = bot.getVelocity();
		
		long time = e.getTime();
		
		double eAbsoluteBearing = bot.getHeadingRadians()
								+ e.getBearingRadians();
		
		double eDistance = e.getDistance();
		myLocation = new Point2D.Double(bot.getX(), bot.getY());
		Point2D enemyLocation = Tools.project(myLocation, eAbsoluteBearing, eDistance);
		
		if (lVelocity == 100) {
			lHeading = eHeading;
			lVelocity = eVelocity;
			
			EnemyData data = new EnemyData();
			data.lastEnergy = e.getEnergy();
			data.location = enemyLocation;
			enemyData.put(e.getName(), data);
			return;
		}
		
		State state = createState(eHeading, eVelocity);
		addState(state);
		
		lHeading = eHeading;
		lVelocity = eVelocity;
		
		//double bPower = bulletPower(e);
		//double bSpeed = Rules.getBulletSpeed(bPower);
		Iterator<Wave> it = waves.iterator();
		while(it.hasNext()) {
			Wave w = it.next();
			if(w.radius(time) > w.distance(myLocation) + 28) {
				it.remove();
			}
		}
		
		/////////////////////////////////////////////
		// Find the deepest Match for the current set
		Point2D best = null;
		
		double eEnergy = e.getEnergy();
		
		double nearest = eDistance;
		Iterator<EnemyData> it2 = enemyData.values().iterator();
		while(it2.hasNext()) {
			double distance = it2.next().location.distance(myLocation);
			if(distance < nearest)
				nearest = distance;
		}
		
		EnemyData tData = enemyData.get(e.getName());
		double delta = 100;
		if(tData != null && eDistance < nearest*1.25)  {
			delta = tData.lastEnergy - eEnergy;
		}
		if(delta >= 0.1 && delta <= 3.0) {
			Wave w = new Wave(enemyLocation);
			w.speed = Rules.getBulletSpeed(delta);
			w.fireTime = time - 1;
			w.myPos = myLocation;
			
			w.linear = myLocation;
			int tickCounter = 0;
			while(w.distance(w.linear) >= w.speed*tickCounter++
				&& battlefield.contains(w.linear)) {
				w.linear = Tools.project(w.linear, lHeading, lVelocity);
			}
			
			waves.add(w);
			
			list.clear();
			
			double bSpeed = Rules.getBulletSpeed(delta);
			
			State tmp = state.lastType;
			int deepest = 0;
			int increment = 0;
			while(tmp != null && ++increment < MAX_MATCH_CHECKS) {
				//Traverse backwards from this, till we run out
				int depth = 0;
				State tmp2 = tmp;
				State current = head;
				while(tmp2.symbol == current.symbol && !tmp2.isBreak) {
					++depth;
					tmp2 = tmp2.last;
					current = current.last;
				}
				
				if(depth > deepest) {
					//Traverse forward till our bullet would intecept!
					Point2D tPos = playItForward(tmp, enemyLocation, myLocation, bSpeed);
					if(tPos != null) {
						deepest = depth;
						best = tPos;
					}
				}
				//Setup for the next pass
				tmp = tmp.lastType;
			}
			w.list = new ArrayList<Point2D>(list);
			w.targetPos = best;
		}
		
		EnemyData data = new EnemyData();
		data.lastEnergy = e.getEnergy();
		data.location = enemyLocation;
		enemyData.put(e.getName(), data);
		
		//Minimal Risk Movement!
		minimalRisk();
	}
	
	double lastBestDanger = Double.POSITIVE_INFINITY;
	public void minimalRisk() {
		Point2D safest = null;
		double lowestDanger = Double.MAX_VALUE;
		
		Point2D center = new Point2D.Double(battlefield.getCenterX(),battlefield.getCenterY());
		double hitDanger = 16;
		double centerDanger = 64;
		
		if(bot.getOthers() == 1) {
			hitDanger = 16;
			centerDanger = 0;
		}
		
		double enemyDanger = 64;
		double wallDanger = 2;
		long time = bot.getTime();
		//double centerDanger = 0.25
		//double absAngle = Tools.absoluteAngle(enemyLocation, myLocation);
		
		for (double angle = 0; angle < Math.PI * 2.0; angle += Math.PI / 32.0) {
			Point2D pos = Tools.project(myLocation, angle, 125 + Math.random()*50);
			double danger = centerDanger/pos.distance(center);
			
			if(myLastLocation != null)
				danger += (enemyDanger*.25) / pos.distance(myLastLocation);
			
			Iterator<EnemyData> it = enemyData.values().iterator();
			while(it.hasNext()) {
				EnemyData data = it.next();
				danger += ((data.lastEnergy/100.0)*enemyDanger) / data.location.distance(pos); 
			}
			for(Wave w : waves) {
				double wDanger = 0;
				if(w.myPos != null) {
					Line2D line = new Line2D.Double(w, w.myPos);
					wDanger += hitDanger / line.ptLineDist(pos);
				}
				if(w.linear != null) {
					Line2D line = new Line2D.Double(w, w.linear);
					wDanger += hitDanger / line.ptLineDist(pos);
				}
				if(w.targetPos != null) {
					Line2D line = new Line2D.Double(w, w.targetPos);
					wDanger += hitDanger / line.ptLineDist(pos);
				}
				double wDist = w.distTo(myLocation, time);
				if(wDist < 100) wDist = 0;
				
				danger += wDanger / Math.max(wDist,1) + 1.0/w.speed;
				//danger += wDanger*10.0/w.radius(time);
			}
			
			/*
				double wDanger = 0;
				if(w.myPos != null) {
					Line2D line = new Line2D.Double(w, w.myPos);
					wDanger += hitDanger / line.ptLineDist(pos);
				}
				if(w.targetPos != null) {
					Line2D line = new Line2D.Double(w, w.targetPos);
					wDanger += hitDanger / line.ptLineDist(pos);
				}
				double wDist = w.distTo(myLocation, time);
				if(wDist < 100) wDist = 0;
				danger += wDanger / Math.max(wDist,1) + 1.0/w.speed;
			 */
			
			danger += wallDanger / Math.max(0, (field.getCenterX() - Math.abs(field.getCenterX() - pos.getX())) - 20);
			danger += wallDanger / Math.max(0, (field.getCenterY() - Math.abs(field.getCenterY() - pos.getY())) - 20);
			
			
			if (!field.contains(pos))
				danger += 100;

			if (danger < lowestDanger) {
				safest = pos;
				lowestDanger = danger;
			}
		}
		if(lowestDanger < lastBestDanger || bot.getOthers() > 1) {
			goTo(safest);
		} else {
			goTo(null);
		}
	}

	Point2D lastGoto = null;

	private void goTo(Point2D destination) {
		if (destination == null) {
			destination = lastGoto;
		}
		lastGoto = destination;
		if (destination == null)
			return;
		double distance = myLocation.distance(destination);
		if(distance < 5) {
			lastBestDanger = Double.POSITIVE_INFINITY;
		}
		double angle = Utils.normalRelativeAngle(Tools.absoluteAngle(myLocation,
				destination) - bot.getHeadingRadians());
		if (Math.abs(angle) > Math.PI / 2) {
			distance = -distance;
			if (angle > 0) {
				angle -= Math.PI;
			} else {
				angle += Math.PI;
			}
		}

		// this is hacked so that the bot doesn't turn once we get to our destination
		bot.setTurnRightRadians(angle * Math.signum(Math.abs((int) distance)));

		// This works better! So we don't bump into the walls! - Chase
		bot.setAhead(Math.cos(angle) * distance);
	}
	
	List<Point2D> list = new ArrayList<Point2D>();
	public final Point2D playItForward(State state, Point2D me, Point2D enemy, double speed) {
		State tmp = state;
		Point2D point = (Point2D)enemy.clone();
		long pTime = -1;
		
		double heading = lHeading;
		double velocity = lVelocity;
		List<Point2D> tmpList = new ArrayList<Point2D>();
		tmpList.add(point);
		while(battlefield.contains(point)
			&& tmp != null && !tmp.isBreak
			&& pTime++ < point.distance(me) / speed) {
			
			heading = Utils.normalRelativeAngle(heading + tmp.angleDelta);
			velocity = Tools.limit(-8,velocity + tmp.velDelta, 8);
			
			point = Tools.project(point, heading, velocity);
			tmpList.add(point);
			tmp = tmp.next;
		}
		if(pTime > point.distance(me) / speed
			&& battlefield.contains(point)) {
			list.clear();
			list.addAll(tmpList);
			return point;
		}
		return null;
	}
	
	public void onBulletHit(BulletHitEvent e) {
		EnemyData data = enemyData.get(e.getName());
		data.lastEnergy -= e.getEnergy();
	}
	
	public final void addState(State s) {
		State tmp = head;
		head = s;
		if(tmp != null)
			tmp.next = head;
		head.last = tmp;
		//Add it to buffer and to the map
		s.lastType = map.get(s.symbol);
		if(s.lastType != null)
			s.lastType.nextType = s;
		map.put(s.symbol, s);
	}
	
	public final void addBreak() {
		State state = new State();
		state.isBreak = true;
		state.symbol = (char)1;
		addState(state);
	}
	
	public final State createState(double heading, double velocity) {
		State state = new State();
		state.angleDelta = Utils.normalRelativeAngle(heading - lHeading);
		state.velDelta = velocity - lVelocity;
		state.symbol = createSymbol(state.angleDelta,state.velDelta); 
		return state;
	}
	
	public static final int createSymbol(double heading, double velocity) {
		int symbol = ((int)(((velocity + 8)/16.0)*64) & 0x3F) << 10;
		symbol |= ((int)(((heading/(Math.PI/18.0))+0.5)*512) & 0x1FF) << 1;
		
		return symbol; 
	}
	
	public void onPaint(Graphics2D g) {
		if (Genesis.isReference) return;
		
		for(Wave w : waves) {
			if(w.list.size() > 0) {
				Iterator<Point2D> it = w.list.iterator();
				g.setColor(Color.WHITE);
				while(it.hasNext()) {
					Point2D pt = it.next();
					g.drawRect((int)pt.getX()-1, (int)pt.getY()-1, 2, 2);
				}
			}
			if(w.targetPos != null) {
				Line2D line = new Line2D.Double(w, w.targetPos);
				g.setColor(Color.RED);
				g.draw(line);
			}
			if(w.myPos != null) {
				Line2D line = new Line2D.Double(w, w.myPos);
				g.setColor(Color.GREEN);
				g.draw(line);
			}
		}
	}
	
	public static class State {
		public State last, next;
		public State lastType, nextType;
		public double angleDelta, velDelta;
		public int symbol;
		boolean isBreak = false;
	}
	
	@SuppressWarnings("serial")
	public static final class Wave extends Point2D.Double {
		double speed = 0;
		long fireTime = 0;
		Point2D targetPos = null;
		Point2D linear = null;
		Point2D myPos = null;
		List<Point2D> list = new ArrayList<Point2D>();
		
		public Wave(Point2D origin) {
			setLocation(origin);
		}
		
		public double distTo(Point2D p, long time) {
			return distance(p) - radius(time);
		}
		
		public double radius(long time) {
			return (time-fireTime)*speed;
		}
	}
}
