package kc.melee;
import kc.melee.*;
import kc.melee.gun.*;
import kc.melee.utils.*;
import robocode.*;
import robocode.util.Utils;
import java.awt.geom.*;

/**
* MeleeManager - Collects the basic game information for Logic.
*/
public class MeleeManager {
	AdvancedRobot robot;
    public MeleeManager(AdvancedRobot robot){this.robot = robot;}
	static final double PI = Math.PI;
	public long gameTime;
	public Point2D.Double myLocation;
	public double myHeading;
	public double myVelocity;
	public double myEnergy;
	public int enemies = 20;
	public int currentEnemies = enemies;
	public int target;
	public int currentScannedEnemy = 0;
	public Enemy[] enemy = new Enemy[enemies];
	public static String[] enemyIndex = new String[20];
	public static int definedEnemies = -1;
	public void init(){
		myLocation = new Point2D.Double(robot.getX(), robot.getY());
		for(int i = 0; i < enemies; i++){
			enemy[i] = new Enemy();
			enemy[i].name = "";
			enemy[i].active = 1; //1 = unknown, 2 = known, 0 = dead
		}
	}
	public void onScannedRobot(ScannedRobotEvent e){
		setVariables();
		double absoluteBearing = Utils.normalAbsoluteAngle(myHeading + e.getBearingRadians());
		double enemyHeading = e.getHeadingRadians();
		double enemyVelocity = e.getVelocity();
		double enemyDistance = e.getDistance();
		double enemyEnergy = e.getEnergy();
		int enemyClockDirection = Math.sin(Utils.normalRelativeAngle(enemyHeading - absoluteBearing)) * enemyVelocity < 0 ? -1 : 1;
		Point2D.Double enemyLocation = KUtils.projectMotion(myLocation, absoluteBearing, enemyDistance);
		boolean known = false;
		for(int i = 0; i < enemies; i++) {
			if(enemy[i].name.equals(e.getName())){
				if(!known){
					currentScannedEnemy = i;
					known = true;
				}
			}
			if(!known){
				if(enemy[i].active == 1){
					enemy[i].name = e.getName();
					enemy[i].location = enemyLocation;
					enemy[i].lastScannedTime = gameTime;
					currentScannedEnemy = i;
					known = true;
				}
			}
		}
		enemy[currentScannedEnemy].clockDirection = enemyClockDirection;
		enemy[currentScannedEnemy].lastLocation = enemy[currentScannedEnemy].location; 	
		enemy[currentScannedEnemy].location = enemyLocation; 			
		enemy[currentScannedEnemy].energy = enemyEnergy;
		enemy[currentScannedEnemy].heading = enemyHeading;
		enemy[currentScannedEnemy].velocity = enemyVelocity;
		enemy[currentScannedEnemy].distance = enemyDistance;
		enemy[currentScannedEnemy].absoluteBearing = absoluteBearing;
		enemy[currentScannedEnemy].active = 2;
		enemy[currentScannedEnemy].lastScannedTime = gameTime;
		if(enemyVelocity != 0){
			if(enemy[currentScannedEnemy].clockDirection != enemyClockDirection){enemy[currentScannedEnemy].lastReverseTime = gameTime;}
			enemy[currentScannedEnemy].clockDirection = enemyClockDirection;
		}
		if(!isDefined(e.getName())){
			definedEnemies++;
			enemyIndex[definedEnemies] = e.getName();
		}
		findTarget();
		setVariables();
	}
	boolean isDefined(String name){
		for(int i = 0; i <= definedEnemies; i++){
			if(name.equals(enemyIndex[i])){return true;}
		}
		return false;
	}
	public int getIndex(String name){
		for(int i = 0; i <= definedEnemies; i++){
			if(name.equals(enemyIndex[i])){return i;}
		}
		return -1;
	}
	public void findTarget(){
		double bestScore = Double.POSITIVE_INFINITY;
		for(int i = 0; i < enemies; i++){
			if(enemy[i].active == 2){
				double score = (target == i ? 1 : 
								(robot.getGunHeat() > 1 ? 1.1 : (robot.getGunHeat() > 0.8 ? 1.4 : (robot.getGunHeat() > 0.5 ? 2 : 5)))) 
								* enemy[i].distance * (enemy[i].energy < 4 * MeleeGun.findBulletPower(enemy[i], this) ? 0.9 : 1);
				if(score < bestScore){
					target = i;
					bestScore = score;
				}
			}
		}
	}
	public double smallestDistance(int robot){
		double smallestDistance = Double.POSITIVE_INFINITY;
		for(int i = 0; i < enemies; i++) {
			if((enemy[i].active == 2)&&(i != robot)){
				double distanceToOther = enemy[robot].location.distance(enemy[i].location);
				if(distanceToOther < smallestDistance){smallestDistance = distanceToOther;}
			}
		}
		if(currentEnemies == 1){smallestDistance = 1;}
		return smallestDistance;
	}
	public Enemy nearestEnemy(){
		int nearest = 0;
		double smallestDistance = Double.POSITIVE_INFINITY;
		for(int i = 0; i < enemies; i++) {
			if(enemy[i].active == 2){
				double distanceToMe = myLocation.distance(enemy[i].location);
				if(distanceToMe < smallestDistance){
					smallestDistance = distanceToMe;
					nearest = i;
				}
			}
		}
		return enemy[nearest];
	}
	public void setVariables(){
		myLocation = new Point2D.Double(robot.getX(), robot.getY());
		myHeading = robot.getHeadingRadians();
		myVelocity = robot.getVelocity();
		myEnergy = robot.getEnergy();
		gameTime = robot.getTime();
		currentEnemies = robot.getOthers();
		for(int i = 0; i < enemies; i++) {
			if(enemy[i].active == 2){
				enemy[i].distance = myLocation.distance(enemy[i].location);
				enemy[i].absoluteBearing = Utils.normalAbsoluteAngle(KUtils.absoluteBearing(myLocation, enemy[i].location));
			}
		}
		if(enemy[target].active != 2){findTarget();}
	}
	public Enemy targetEnemy(){
		return enemy[target];
	}
	public Enemy findEnemy(String name){
		for(int i = 0; i < enemies; i++) {
			if(name.equals(enemy[i].name)){return enemy[i];}
		}
		return enemy[0];
	}
	public void onRobotDeath(RobotDeathEvent e) {
        for(int i = 0; i < enemies; i++) {
			if(enemy[i].name.equals(e.getName())){enemy[i].active = 0;}
		}
		setVariables();
	}
}
