package kc.melee.movement;
import kc.melee.*;
import kc.melee.utils.*;
import robocode.*;
import robocode.util.Utils;
import java.io.*;
import java.awt.geom.*;
import java.awt.Color;

/**
 * DuelMovement - Operates Logic's one - on - one movement.
 */
public class DuelMovement {
	AdvancedRobot robot;
	MeleeManager manager;
    public DuelMovement(AdvancedRobot robot, MeleeManager manager) {
		this.robot = robot;
		this.manager = manager;
	}
	static final double PI = Math.PI;
	double enemyEnergy;
	double lastEnemyBulletPower = 2.0;
	int myDirection = 1;
	int movementMode = 0;
	int stopNGoHits = 0;
	public void onScannedRobot(ScannedRobotEvent e){
		double enemyDistance = e.getDistance();
		if ((enemyEnergy -= e.getEnergy()) >= 0.1 && enemyEnergy <= 3.0){
			lastEnemyBulletPower = enemyEnergy;
			if(movementMode == 0){
				robot.setMaxVelocity(Math.abs(robot.getVelocity()) < 2 ? 8 : 1);
				robot.setAhead(Double.POSITIVE_INFINITY * myDirection);
			}
		}
		double bulletImpactTime = enemyDistance / KUtils.bulletSpeed(lastEnemyBulletPower);
		if(Math.random() < (bulletImpactTime * bulletImpactTime * 0.00003232429335) - (bulletImpactTime * 0.004619996) + 0.20680504 && movementMode == 1){
			robot.setAhead(1000 * (myDirection = -myDirection));
			robot.setMaxVelocity(8);
		}
		double evasion = (0.0000000000030748 * Math.pow(enemyDistance, 4)) - (0.0000000078308 * Math.pow(enemyDistance, 3)) + (0.0000057413 * Math.pow(enemyDistance, 2)) + (0.000207 * enemyDistance) - 0.89294;
		double maxSmoothing = PI/3;
		if(enemyDistance < 600){maxSmoothing = PI/3.5;}
		if(enemyDistance < 350){maxSmoothing = PI/5;}
		double targetHeading = manager.enemy[manager.target].absoluteBearing + PI/2;
		double oldHeading = targetHeading;
		Rectangle2D battleField = KUtils.makeField(robot.getBattleFieldWidth(), robot.getBattleFieldHeight(), 30.0);
		while (!battleField.contains(KUtils.projectMotion(manager.myLocation, (targetHeading -= myDirection * 0.03), (movementMode == 0 ? 100.0 : 170.0) * myDirection)));
		if(Math.abs(Utils.normalRelativeAngle(targetHeading - oldHeading)) > maxSmoothing){
			targetHeading = oldHeading;
			robot.setAhead(movementMode * 1000 * (myDirection = -myDirection));
		}	
		robot.setTurnRightRadians(Utils.normalRelativeAngle(targetHeading - manager.myHeading));
		enemyEnergy = e.getEnergy();
	}
	public void onHitByBullet(HitByBulletEvent e){
		enemyEnergy += 3 * e.getPower();
		if(++stopNGoHits > 1 || manager.myEnergy < 15){movementMode = 1;}
	}
	public void onBulletHit(BulletHitEvent e){
		double power = e.getBullet().getPower();
		enemyEnergy -= (4 * power) + (2 * Math.max(power - 1, 0));
	}
}
