package kc.melee.movement;
import kc.melee.*;
import kc.melee.utils.*;
import robocode.*;
import robocode.util.Utils;
import java.io.*;
import java.awt.geom.*;
import java.awt.Color;

/**
 * MeleeMovment - Controls how Logic moves in melee battles
 */
public class MeleeMovement {
	AdvancedRobot robot;
	MeleeManager manager;
    public MeleeMovement(AdvancedRobot robot, MeleeManager manager) {
		this.robot = robot;
		this.manager = manager;
	}
	static final double PI = Math.PI;
	static final double WALL_MARGIN = 30;
	static Rectangle2D battleField;
	Point2D.Double nextLocation;
	Point2D.Double lastLocation;
	Point2D.Double bestLocation;
	int push = 0;
	public void init(){
		nextLocation = lastLocation = bestLocation = manager.myLocation;
		battleField = KUtils.makeField(robot.getBattleFieldWidth(), robot.getBattleFieldHeight(), WALL_MARGIN);
	}
	public void move(){
		if(Math.random() < 0.25 && manager.currentEnemies == 2){push = 1;}else{push = 0;}
		double leastRisk = Double.POSITIVE_INFINITY;
		for(int i = 0; i < (Math.abs(robot.getDistanceRemaining()) < 10 ? 150 : 50); i++){
			double heading = Math.random() * PI * 2;
			double distance = Math.min((Math.random() * 200) + 100, manager.nearestEnemy().distance * 0.8);
			Point2D.Double testLocation = KUtils.projectMotion(manager.myLocation, heading, distance);
			double risk = evaluatePosition(testLocation, heading);
			if(risk < leastRisk){
				leastRisk = risk;
				bestLocation = testLocation;
			}
		}
		if(Math.abs(robot.getDistanceRemaining()) < 10 || (manager.currentEnemies != 2 && leastRisk < 0.75 * evaluatePosition(nextLocation, KUtils.absoluteBearing(manager.myLocation, nextLocation)))){
			driveTo(bestLocation);
			lastLocation = manager.myLocation;
			nextLocation = bestLocation;
		}
		robot.setMaxVelocity(Math.abs(robot.getTurnRemaining()) < 15 ? 8 : 8 - Math.min(Math.abs(robot.getTurnRemaining()) / 8, 8));
	}
	double evaluatePosition(Point2D.Double location, double heading){
		if(!battleField.contains(location)){return Double.POSITIVE_INFINITY;}
		double risk = Math.sqrt(manager.currentEnemies) / ((push == 1 ? 15 : 25) * location.distanceSq(lastLocation));
		for(int i = 0; i < manager.enemies; i++){
			if(manager.enemy[i].active == 2){
				risk += Math.min(manager.enemy[i].energy / manager.myEnergy, 2) * (location.distance(manager.enemy[i].location) < manager.smallestDistance(i) ? (manager.currentEnemies == 3 ? 6 : 3) : 1) * (location.distance(manager.enemy[i].location) < 50.0 ? 5 : 1) / Math.pow(location.distance(manager.enemy[i].location), 2);
				risk += Math.min(manager.enemy[i].energy / manager.myEnergy, 2) * (manager.myLocation.distance(manager.enemy[i].location) < manager.smallestDistance(i) ? 2 : 1) * Math.abs(Math.cos(Utils.normalRelativeAngle(manager.enemy[i].absoluteBearing - heading))) / Math.pow(location.distance(manager.enemy[i].location), 2);
			}
		}
		return risk;
	}
	public void driveTo(Point2D.Double location){
		double turn = Utils.normalRelativeAngle(KUtils.absoluteBearing(manager.myLocation, location) - manager.myHeading);
		double distance = manager.myLocation.distance(location);
		int myDirection = 1;
		if(turn < -PI/2){
			turn += PI;
			myDirection = -1;
		}
		if(turn > PI/2){
			turn -= PI;
			myDirection = -1;
		}
		robot.setTurnRightRadians(turn);
		robot.setAhead(myDirection * distance);
	}
}
