package kc.mini;
import robocode.*;
import robocode.util.Utils;
import java.awt.geom.*;
import java.awt.Color;

/**
* Vyper, by Kevin Clark (Kev).
* Mini bot with multimode movement and guess factor targeting. 
* See http://robowiki.net/cgi-bin/robowiki?Vyper for more information. 
*/
public class Vyper extends AdvancedRobot {
	static final int AIM_FACTORS = 33;
	static final int MIDDLE_FACTOR = 16;
	static Point2D.Double enemyLocation;
	
	static double enemyEnergy;
	static double lastEnemyBulletSpeed;
	static int enemyClockDirection = 1;
	double lastVelocity;
	int timeSinceAccel;
	
	static double moveFactor = 10000;
	static int movementMode;
	int modeHits;
	
	static double[][][][][] dataFast = new double[3][3][3][2][AIM_FACTORS];
	static double[][][][][][] dataSlow = new double[5][4][5][3][2][AIM_FACTORS];
	
	public void run() {
		setColors(Color.blue, Color.blue, Color.yellow);
		setAdjustRadarForGunTurn(true);
        setAdjustGunForRobotTurn(true);
		do {
	   	 	turnRadarRightRadians(Double.POSITIVE_INFINITY); 
		} while (true);
	}
	
	public void onScannedRobot(ScannedRobotEvent e) {
		if ((enemyEnergy -= e.getEnergy()) >= 0.1 && enemyEnergy <= 3.0) {
			lastEnemyBulletSpeed = 20 - (3 * enemyEnergy);
			if(movementMode == 1) {
				setAhead(Math.abs(getDistanceRemaining()) > 0 ? 0 : moveFactor);
			}
		}
		
		Wave w = new Wave();
		
		double absoluteBearing = w.absoluteBearing =  e.getBearingRadians() + getHeadingRadians(); 
		Point2D.Double myLocation = w.location = new Point2D.Double(getX(), getY());
		enemyLocation = projectMotion(myLocation, absoluteBearing, e.getDistance());
		
		if(movementMode != 1) {
			if(Math.random() / movementMode < 0.325 * Math.pow(e.getDistance() / lastEnemyBulletSpeed, -0.65)) {
			//if(movementMode == 2 && Math.random() < ((enemyEnergy =  e.getDistance() / lastEnemyBulletSpeed) * ((enemyEnergy * 0.00003232429335) - 0.004619996)) + 0.20680504) {
				moveFactor = -moveFactor;
			}
			setAhead(moveFactor);
		}
	
		Rectangle2D.Double battleField = new Rectangle2D.Double(50, 50, 700, 500);
		if(movementMode != 1) {
			battleField = new Rectangle2D.Double(25, 25, 750, 550);
		}
	
		double targetHeading;
		double smoothedHeading = targetHeading = absoluteBearing + Math.PI/2 - (moveFactor * (e.getDistance() - 510) * (2 + movementMode) / 22000000);
		while(!rectangleContains(battleField, myLocation, (smoothedHeading -= moveFactor * 0.000003), (movementMode == 1 ? 0.01 : 0.017) * moveFactor));
		if((((enemyEnergy = Math.abs(targetHeading - smoothedHeading)) > Math.PI/3.2 && e.getDistance() > 350)||(enemyEnergy > Math.PI/4.5))&&(rectangleContains(battleField, myLocation, targetHeading, 0.004 * moveFactor) || enemyEnergy > Math.PI/2.5)) {
			setAhead(Math.abs(movementMode - 1) * (moveFactor = -moveFactor));
		}
		setTurnRightRadians(Utils.normalRelativeAngle(smoothedHeading - getHeadingRadians()));
		
		double bulletPower = e.getDistance() < 125 ? 3 : 1.9;
		if(getEnergy() < 20 && (e.getEnergy() > getEnergy() || e.getEnergy() > 8)) {
			bulletPower = 1.4;
		}
		bulletPower = Math.min(Math.min(bulletPower, e.getEnergy() / 4), getEnergy() / 5);
		
		if(e.getVelocity() != 0) {
			enemyClockDirection = Math.sin(e.getHeadingRadians() - absoluteBearing) * e.getVelocity() < 0 ? -1 : 1;
		}
		w.clockDirection = enemyClockDirection * Math.asin(8 / (w.speed = 20 - (3 * bulletPower)));
		w.weight = 1;
		
		int bestGF = 4;
		if(Math.abs(enemyEnergy = (Math.abs(e.getVelocity()) - lastVelocity)) > 0.1) {
			timeSinceAccel = 0;
			bestGF = enemyEnergy > 0 ? 0 : 1;
		} else {
			if((enemyEnergy = ++timeSinceAccel / (e.getDistance() / w.speed)) < 0.7) {
				bestGF = 3;
			}
			if(enemyEnergy < 0.3) {
				bestGF = 2;
			}
		}
		
		battleField = new Rectangle2D.Double(17.999, 17.999, 764, 564);
			
		w.dataFast = dataFast[Math.min((int)((e.getDistance() / w.speed) / 18.65), 2)]
		             [(int)((lastVelocity = Math.abs(e.getVelocity())) / 2.67)]
				     [Math.min(bestGF, 2)]
					 [rectangleContains(battleField, myLocation, absoluteBearing + 0.6 * enemyClockDirection, e.getDistance()) ? 0 : 1];
		w.dataSlow = dataSlow[Math.min((int)((e.getDistance() / w.speed) / 11.19), 4)]
		             [(int)((lastVelocity = Math.abs(e.getVelocity())) / 2.01)]
				     [bestGF]
					 [rectangleContains(battleField, myLocation, absoluteBearing + 0.6 * enemyClockDirection, e.getDistance()) ? 0 : (rectangleContains(battleField, myLocation, absoluteBearing + 0.3 * enemyClockDirection, e.getDistance()) ? 1 : 2)]
					 [rectangleContains(battleField, myLocation, absoluteBearing - 0.5 * enemyClockDirection, e.getDistance()) ? 0 : 1];
		
		
		bestGF = MIDDLE_FACTOR;
		int gf = AIM_FACTORS - 1;
		enemyEnergy = 0;
		do {
			if((smoothedHeading = w.dataFast[gf] + Math.pow(w.dataSlow[gf], 1.2)) > enemyEnergy) {
				bestGF = gf;
				enemyEnergy = smoothedHeading;
			}
		} while(gf-- > 0);
		setTurnGunRightRadians(Utils.normalRelativeAngle(absoluteBearing - getGunHeadingRadians() + ((bestGF - MIDDLE_FACTOR) * w.clockDirection / MIDDLE_FACTOR)));
		
		if((getEnergy() - bulletPower > 1 || e.getDistance() < 150 || enemyEnergy == 0)) {
			if(getGunHeat() == 0.0) {
				setFire(bulletPower);
				w.weight = 3;
			}
			addCustomEvent(w);
		}
		
		setTurnRadarRightRadians(Utils.normalRelativeAngle(absoluteBearing - getRadarHeadingRadians()) * 2);
		
		enemyEnergy = e.getEnergy();
	}
	
	public void onHitByBullet(HitByBulletEvent e) {
		enemyEnergy += e.getPower() * 3;
		if(movementMode != 2 && getOthers() * ++modeHits > getRoundNum() + 2 + movementMode) {
			movementMode++;
			setAhead(0);
			modeHits = 0;
		}	
	}
	
	public void onBulletHit(BulletHitEvent e) {
		enemyEnergy -= (4 * e.getBullet().getPower()) + 2 * Math.max(e.getBullet().getPower() - 1, 0); 
	}
	
	public void onDeath(DeathEvent e) {
		movementMode = Math.min(movementMode + 1, 2);
	}
	
	static boolean rectangleContains(Rectangle2D.Double r, Point2D.Double location, double heading, double distance) {
		return r.contains(projectMotion(location, heading, distance));
	}
	
	static Point2D.Double projectMotion(Point2D.Double location, double heading, double distance) {	
		return new Point2D.Double(location.x + (distance * Math.sin(heading)), location.y + (distance * Math.cos(heading)));			
	}
	
	public class Wave extends Condition {
		double[] dataFast;
		double[] dataSlow;
		Point2D.Double location;
		double absoluteBearing;
		double clockDirection;
		double radius;
		double speed;
		double weight;
		
		public boolean test() {
			if(enemyLocation.distance(location) <= (radius += speed) + speed && getOthers() > 0) {
				int gf = AIM_FACTORS - 1;
				do {
					dataFast[gf] *= 0.998;
					dataSlow[gf] *= 0.998;
					dataFast[gf] += (radius = (weight * Math.pow(0.3, Math.abs(gf - (int)(MIDDLE_FACTOR + Math.round(MIDDLE_FACTOR * Utils.normalRelativeAngle(Math.atan2(enemyLocation.x - location.x, enemyLocation.y - location.y) - absoluteBearing) / clockDirection))))));
					dataSlow[gf] += radius;
				} while(gf-- > 0);
				removeCustomEvent(this);
			}
			return false;
		}
	}
}
