package nz.jdc.micro;

// import java.awt.Color;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import robocode.AdvancedRobot;
import robocode.Condition;
import robocode.HitByBulletEvent;
import robocode.Rules;
import robocode.ScannedRobotEvent;
import robocode.util.Utils;

/**
 * HedgehogGF 
 * Named by my wife, Yayoi, because hedgehogs are cute. Hopefully this one has some spikes.
 * My fifth bot. Second micro bot.
 * 
 * Firing: Guess Factor Targeting.
 * Movement: Bullet dodge, or flat.
 * 
 * Hmmm. Pretty standard microbot, couldn't find much original to add. Too many cool ideas
 * proved to be flops :(
 * 
 * Credits: 
 * Gun based on Thorn (by Kevin Clark) and RaikoMicro (by Jamougha).
 * Movement is inspired by Waylander (by JK) and Weekend Obsession (by Eric Simonton).
 * Lots of other stuff inspired by or borrowed from all the good articles on the wiki.
 *  
 * Version 1.0: 
 *   Codesize : 748
 *   
 * Version 1.1:
 *   No code change, but had to fix a build bug and redeploy a new version  
 *   Codesize : 748
 *   
 * Version 1.2:
 *   Found HedgehogGF to be weak against rammers, so had to add an increase of
 *   bullet power at short range. Unfortunately can no longer fit colour.  
 *   codesize : 749
 *
 * Version 1.3:
 *   Tuned the stop and go movement.
 *   Codesize : ???
 *
 * @author John Cleland 
 */
public class HedgehogGF extends AdvancedRobot {

  // for movement -------------------------------------------------------------
  private final static double PERPENDICULAR    = Math.PI / 2.0;
  private final static double MOVE_DISTANCE    = 55;
  private final static double WALL_STICK       = 160.0;
  private final static double SMOOTH_INC       = 0.07;
  
  private final static double CLOSE_FCT        = 400;

  private final static double PREFERRED_RANGE  = 500;
  private final static double RM_MULT          = 18;
  private final static double REVERSE_ANGLE    = Math.PI / 4.0;
  
  //.15 is hit rate, .1 is gun cool rate, 2.0 is a very approximate bullet power to gun heat conversion
  private final static double HIT_RATE_SWITCH  = 0.17 * 0.1 * 2.0;

  // flattener
  private final static double REVERSE_CHANCE   = 0.6;
  private final static double REVERSE_DIST_ADJ = 150;
  private final static double REVERSE_SUB      = 0.02;

  private static double  direction             = 1;
  private static double  enemyBulletVelocity   = 15.0;
  private static int     hitRate;
  private static boolean flat;
  
  private static double  enemyEnergy;
  
  // for gun ------------------------------------------------------------------
  
  // constants
  private final static double   BULLET_POWER    = 1.9;
  private final static double   BULLET_VELOCITY = 20 - 3 * BULLET_POWER;
  private final static double   EFFECTIVE_GF0   = 15;     // \ the difference between these two gives us some bin leeway      
  private final static int      MIDDLE_FACTOR   = 25;     // / so we don't get out of bounds exceptions
  private final static int      AIM_FACTORS     = MIDDLE_FACTOR * 2 + 1;
  private final static int      AIM_FACTORS_1   = AIM_FACTORS - 1;
  private final static double   ANGLE_TO_BIN    = 0.0396; // asin(8 / BULLET_VELOCITY) / EFFECTIVE_GF0;
  private final static int      DIST_DIVISOR    = 140;
  
  // segments
  private final static int      NSEG_DIST       = 7;      // DIST_DIVISOR * NSEG_DIST > 950
  private final static int      NSEG_ACC        = 3;
  private final static int      NSEG_VEL        = 9;     
  private final static int      NSEG_WALL       = 2;
    
  // attributes
  private static Point2D.Double enemyLocation;
  private static int            lastVel;
  private static double         eoDir = 1;
  private static int[][][][][]  data = new int[NSEG_DIST][NSEG_ACC][NSEG_VEL][NSEG_WALL][AIM_FACTORS];
  
  //---------------------------------------------------------------------------

  /**
   * Bot run method.
   */
  public void run() {
    // Yayoi says he has to be pink :)
    // but unfortunately I ran out of space :(
    // setBodyColor(Color.PINK);
    
    // helps targeting accuracy
    setAdjustGunForRobotTurn(true);
    
    // necessary for good radar lock
    setAdjustRadarForGunTurn(true);
    
    // change movement mode if we have exceeded our threshold
    if (hitRate > 0) {
      flat = true;
    }

    // start our infinite radar loop
    while (true) {
      turnRadarRightRadians(1);
    }
  }

  /**
   * Scan handler, where it all happens. 
   */
  public void onScannedRobot(ScannedRobotEvent r2e)
  {
    // register vars
    int    r3i;
    double r4d;
    
    // other vars
    Wave           wave = new Wave();
    Point2D.Double myLocation;
    double         absBearing;
    double         headingRadians;
    double         eDistance;
    double         smoothedDir;
    double         lClockDirection;
    double         latV;
    double         vel;
    int[]          wgf;
    int            bestGF = MIDDLE_FACTOR;
    
    // common and radar -------------------------------------------------------

    // bearing to enemy, location etc
    myLocation = wave.location = new Point2D.Double(getX(), getY());
    // r4d is absBearing
    r4d = wave.absoluteBearing = absBearing = r2e.getBearingRadians()+ (headingRadians = getHeadingRadians());
    enemyLocation = project(myLocation, r4d, (eDistance = r2e.getDistance()));

    // battlefield
    Rectangle2D.Double field = new Rectangle2D.Double(18, 18, 764, 564);

    // do radar scan
    setTurnRadarRightRadians(2.0 * Utils.normalRelativeAngle(r4d - getRadarHeadingRadians()));
    
    // calc enemy latv and orbit direction
    if ((latV = (Math.sin(r2e.getHeadingRadians() - r4d) * (vel = r2e.getVelocity()))) != 0) {
      eoDir = Math.signum(latV);
    }

    // movement ---------------------------------------------------------------
    
    // update hit stats
    hitRate--;
    
    // choose preferred direction, perpendicular + range control
    // r4d is now turn offset from prerpendicular
    r4d = PERPENDICULAR - (eDistance - PREFERRED_RANGE) / CLOSE_FCT;
    
    // wall smoothing
    while(!field.contains(project(myLocation, smoothedDir = absBearing + direction * (r4d -= SMOOTH_INC), WALL_STICK)));
    
    // flat movement reverse chance
    if((flat && (Math.random() < REVERSE_CHANCE * Math.sqrt(enemyBulletVelocity / (eDistance + REVERSE_DIST_ADJ)) - REVERSE_SUB)) || r4d < REVERSE_ANGLE ) {
        direction = -direction;
    }
    
    // back as front turn and random velocity change
    setTurnRightRadians(Math.tan(smoothedDir -= headingRadians));
    setMaxVelocity(7 + Math.random() * RM_MULT);

    // move if flat or bullet detected, with back as front direction control
    if (((r4d = enemyEnergy - (enemyEnergy = r2e.getEnergy())) < 3.01 && r4d > 0.09) || flat) {
      setAhead((MOVE_DISTANCE - Rules.getBulletSpeed(r4d)) * Math.signum(Math.cos(smoothedDir)));
    }
    
    
    // gun --------------------------------------------------------------------
    
    wave.clockDirection = lClockDirection = eoDir * ANGLE_TO_BIN; 
    
    // choose segments
    wgf = wave.guessFactors = data
      /* dist   */ [((int) eDistance) / DIST_DIVISOR]
      /* acc    */ [Integer.signum(lastVel - (lastVel = Math.abs((int) vel))) + 1]
      /* vel    */ [(int) Math.abs(latV)]
      /* wall   */ [field.contains(project(myLocation, absBearing + (EFFECTIVE_GF0 * lClockDirection), eDistance)) ? 0 : 1];
    
    // aim, always the default GF0 if enemy energy is zero
    if (enemyEnergy > 0.0) {
      r3i = AIM_FACTORS_1;
      do {
        if (wgf[r3i] > wgf[bestGF]) {
          bestGF = r3i;
        }
      } while(r3i-- > 0);
    } 

    // add wave 
    addCustomEvent(wave);

    // fire bullet if we have enough energy
    if (((r4d = getEnergy()) > 4.0)) {
      setFireBullet(BULLET_POWER + (int) (120 / ((int) eDistance)));
    }
    
    // turn gun
    setTurnGunRightRadians(Utils.normalRelativeAngle(absBearing - getGunHeadingRadians() + ((bestGF - MIDDLE_FACTOR) * lClockDirection)));
  }
  
  /** */
  public void onHitByBullet(HitByBulletEvent e){
    enemyBulletVelocity = e.getVelocity();  
    hitRate += e.getPower() / HIT_RATE_SWITCH;
  } 

  /** project a point */
  private static Point2D.Double project(Point2D.Double location, double angle, double distance){
    return new Point2D.Double(location.x + distance * Math.sin(angle), location.y + distance * Math.cos(angle));
  }

  // Wave inner class ---------------------------------------------------------
  public class Wave extends Condition {
    Point2D.Double location;
    int[]  guessFactors;
    double clockDirection;
    double absoluteBearing; 
    double distance;
    
    public boolean test() {
      if (enemyLocation.distance(location) <= (distance += BULLET_VELOCITY)) { 
          guessFactors[MIDDLE_FACTOR + (int)(Utils.normalRelativeAngle(Math.atan2(enemyLocation.x - location.x, enemyLocation.y - location.y) - absoluteBearing) / clockDirection)]++;
        removeCustomEvent(this);
      }
      return false;
    }
  }
}
