package nz.jdc.nano;

import java.awt.Color;

import robocode.AdvancedRobot;
import robocode.HitByBulletEvent;
import robocode.HitWallEvent;
import robocode.ScannedRobotEvent;
import robocode.util.Utils;

/**
 * My second nano bot. Close sibling to NeophyteSRAL.
 * Firing: simple prediction with rolling average linear velocity and linear targeting.
 * Movement:  Bullet dodge. Toggles between uni-directional and reversing moves.
 * 
 * Credits: Pretty much everybody, especially those who contributed helpful articles to the
 * wiki, particularly the stuff on flat movement, rolling averages and nano-linear targetting.
 * Movement is based on the very elegant movement strategy of dft.Freddie 1.32 by Alcatraz271.
 *  
 * Version 1.0: Initial version. Hopefully not too sucky. Unfortunately testing shows it is
 *   weak against the plethora of pattern gun / random movement bots out there.
 *   Codesize 248.
 *
 * Version 1.1: Minor tweaks to a few params.
 *   Codesize 249.
 *   
 * Version 1.2: Change movement to one derived from dft.Freddie by Alcatraz271.  
 *   Codesize 249.
 *
 * @author John Cleland
 */
public class NeophytePRAL extends AdvancedRobot {
  // constants
  // tunable constants
  // for movement
  private final static double MOVE_DISTANCE        = 250.0;
  private final static double NON_RAND             = 0.15;
  private final static double PREFERRED_RANGE      = 180.0;
  private final static double CLOSE_FCT            = 950 * MOVE_DISTANCE;
 
  // for gun
  private final static double ROLL_FACTOR         = 8;
  private final static int    P_OFFS              = 8;
  private final static int    P_BINS              = 1 + 2 * P_OFFS;
  private final static double BULLET_POWER_FACTOR = 500.0D;
  private final static double BULLET_SPEED        = 14.5; 
  
  // attributes
  // for movement
  private static double enemyEnergy;
  private static double dir                       = MOVE_DISTANCE;
  private static int    type                      = 1;
  
  // for gun  
  private static double   rAvgCurrent;
  private static int      lvIndexPrevious;
  private static double[] rAvgLateralVelocity = new double[P_BINS];


  /**
   * Bot run method.
   */
  public void run() {
    // finally have space for colour :)
    setBodyColor(Color.WHITE);
    
    // helps targeting accuracy
    setAdjustGunForRobotTurn(true);

    // start our infinite radar loop
    turnRadarRightRadians(Double.POSITIVE_INFINITY);
  }
  
  /**
   * Scan handler.
   * This is where it all happens.
   */
  public void onScannedRobot(ScannedRobotEvent e) {
    // register vars, should be the most commonly used
    int    lvIndexCurrent;
    double rd;

    // other vars
    double enemyDistance;
    
    // radar scan
    setTurnRadarLeftRadians(getRadarTurnRemaining());
    
    // bullet dodge a random distance
    if ((rd = enemyEnergy - (enemyEnergy = e.getEnergy())) <= 3.0 && rd > 0.0) {
      setAhead((dir *= (double) type) * (Math.random() + NON_RAND));
    }
    
    // turn perpendicular with distance control
    setTurnRightRadians(Math.cos((rd = e.getBearingRadians()) - ((enemyDistance = e.getDistance()) - PREFERRED_RANGE) * dir / CLOSE_FCT));
    // rd has enemy relative bearing
    
    // targetting
    // update our rolling average lateral velocity and get an integer index for predictive lookup
    // rd is updated to enemy absolute bearing
    // as usual a nasty compound statement with nested assignments to save space
    lvIndexCurrent = (int) (rAvgCurrent = (rAvgCurrent * ROLL_FACTOR + (e.getVelocity() * Math.sin(e.getHeadingRadians() - (rd += getHeadingRadians())))) / (ROLL_FACTOR + 1.0)) + P_OFFS;
    
    // nano linear targetting fn
    setTurnGunRightRadians(Utils.normalRelativeAngle(rd - getGunHeadingRadians()
      + (rAvgLateralVelocity[lvIndexCurrent] / BULLET_SPEED)));
   
    // firing and stat update
    if (setFireBullet(BULLET_POWER_FACTOR / enemyDistance) != null) {
      rAvgLateralVelocity[lvIndexPrevious] = rAvgCurrent;
      lvIndexPrevious = lvIndexCurrent;
    }
  } 
  
  /**
   * Hit a wall, reverse direction and start moving immediately.
   */
  public void onHitWall(HitWallEvent e) {
    setAhead(dir = -dir);
  }

  /**
   * Hit by bullet, change movement strategy.
   */
  public void onHitByBullet(HitByBulletEvent e) {
    type = -type;
  }
}
