package sgp;
import robocode.*;
import java.util.*;

/**
 * FuturePositionArray - a class by Simon Parker
 */

public class FuturePositionArray
{
	private PatternAnalyser analyser;
	static protected int MAX_LOOKAHEAD_TIME = 150;
	final private int MAX_VELOCITY = 8;
	final private int NUM_VELOCITIES = MAX_VELOCITY * 2 + 1;


	private int[] updateCountArray = new int[NUM_VELOCITIES];
	private Coordinate[][] positionArray = new Coordinate[NUM_VELOCITIES][MAX_LOOKAHEAD_TIME];

	public FuturePositionArray(PatternAnalyser patternAnalyser)
	{
		analyser = patternAnalyser;
		for (int i = 0; i < NUM_VELOCITIES; i++)
		{
			for (int j = 0; j < MAX_LOOKAHEAD_TIME; j++)
			{
				positionArray[i][j] = new Coordinate();
			}
		}
	}

	public Coordinate getPosition(double velocity, double time)
	{
		int velocityIndex = getVelocityIndex(velocity);
		int floorTime = (int)Math.floor(time);
		int ceilingTime = (int)Math.ceil(time);

		if ((floorTime >= MAX_LOOKAHEAD_TIME) || (ceilingTime >= MAX_LOOKAHEAD_TIME))
		{
			floorTime = MAX_LOOKAHEAD_TIME - 2;
			ceilingTime = floorTime + 1;

		}

		if ((floorTime < 0) || (ceilingTime < 0))
		{
			floorTime = 0;
			ceilingTime = 1;
		}


//???		JollyNinja.getInstance().out.println("t0 " + floorTime + ", t1 " + ceilingTime + ", v " + velocityIndex);

		double dX = positionArray[velocityIndex][ceilingTime].x - positionArray[velocityIndex][floorTime].x;
		double dY = positionArray[velocityIndex][ceilingTime].y - positionArray[velocityIndex][floorTime].y;
		double x = positionArray[velocityIndex][floorTime].x + dX * (time - (double)floorTime);
		double y = positionArray[velocityIndex][floorTime].y + dY * (time - (double)floorTime);

		return new Coordinate(x, y);
	}

	public void update(Coordinate position, double velocity)
	{
		int velocityIndex = getVelocityIndex(velocity);

		updateCountArray[velocityIndex]++;

		for (int lookaheadTime = 0; lookaheadTime < MAX_LOOKAHEAD_TIME; lookaheadTime++)
		{
			//calculate the normalised delta position for each future time
			PositionSample positionSample = analyser.getPositionSample(lookaheadTime);
			if (positionSample == null) continue;

			Coordinate latestPosition = position.minus(positionSample.position).rotate(-positionSample.heading_deg);

			//cumulative average position
			double N = (double)updateCountArray[velocityIndex]; //number of samples that the average has been calculated over
			N = Math.min(N, 10);

		//	positionArray[velocityIndex][lookaheadTime].set(latestPosition);

			positionArray[velocityIndex][lookaheadTime].set
			(
				((N - 1.0) * positionArray[velocityIndex][lookaheadTime].x + latestPosition.x) / N,
				((N - 1.0) * positionArray[velocityIndex][lookaheadTime].y + latestPosition.y) / N
			);

		}
	}

	private int getVelocityIndex(double velocity)
	{
		return (int)Math.round(velocity) + MAX_VELOCITY;
	}
}